open Unify

let rec repl ch ctx env =
  let _ = Printf.printf "# %!" in
  let lexbuf = Lexing.from_channel ch in
  try
    let d = Parser.decl Lexer.token lexbuf in
    let (ty, s) = Typecheck.typecheck_decl ctx d in
    let ctx' =
      sub_ctx s (Context.ctx_add ctx (Ast.var_of_decl d) (gen ctx ty))
    in
    let (x, v) = Interp.interp_decl d env in
    let env' = Interp.env_add env x v in
    let _ = Format.fprintf Format.std_formatter
              "%s: %s = %a\n"
              (Ast.var_of_decl d)
              (string_of_sch (gen ctx ty))
              Print.pprint_value v
    in
    let _ = Format.print_flush () in
    repl ch ctx' env'
  with
     | Typecheck.TypeError s ->
        (Printf.printf "Type error: %s\n%!" s;
         repl ch ctx env)
     | Unify.UnificationFailure s ->
        (Printf.printf "Type error: %s\n%!" s;
		       repl ch ctx env)
     | Ast.SyntaxError -> repl ch ctx env
     | Lexer.Quit -> exit 0
     | Parser.Error ->
        (Printf.printf "Syntax error\n%!";
         repl ch ctx env)
;;

if Array.length Sys.argv > 2 then
  (Printf.printf "Usage: ./miniml [filename]\n";
   exit 1)
else if Array.length Sys.argv = 1 then
  (Printf.printf "Welcome to MiniCaml. Type #quit;; to exit.\n%!";
   repl stdin Context.empty_ctx Interp.empty_env)
else
  let fname = Array.get Sys.argv 1 in

  let chan = open_in fname in
  let lexbuf = Lexing.from_channel chan
  in
  let p = Parser.prog Lexer.token lexbuf in
  let _ = Typecheck.tc_prog p in
  let _ = Printf.printf "\n" in
  ignore (Interp.interp_prog p (Interp.empty_env))
